from typing import Optional, Tuple

import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.axes import Axes

from ifa.math import sigmoid, logit
from ifa.segmentor import *
from ifa.ci import *


def plot_result(result: pd.DataFrame, x_col: Optional[str]=None,
                y_axis: str='bias',
                figsize: Tuple[int, int]=(10, 4),
                width_ratios: Tuple[int, int]=(1, 10),
                color = 'blue',
                ax: Optional[np.ndarray]=None):
    if ax is None:
        _, ax = plt.subplots(1,2, figsize=figsize, gridspec_kw={'width_ratios': [*width_ratios]}, sharey=True)
    ax = ax.ravel()

    _y_axis = result['bias']
    ymin = result['bias_lower']
    ymax = result['bias_upper']
    if y_axis == 'bias':
        pass
    elif y_axis == 'mape':
        def _mape(key):
            s = np.abs(result['p_q50'] - sigmoid(logit(result['p_q50']) + result[key]))
            s /= result['p_q50']
            s = s.mean()
            return s
        _y_axis = _mape('bias')
        ymin = _mape('bias_lower')
        ymax = _mape('bias_upper')     
    else:
        raise Exception(f'unrecognized option {y_axis=}')

    null_msk = result['i'] == -1
    _n_null = _y_axis[null_msk].size

    # x null bias plot
    ax[0].axhline(0, linestyle='--', color='gray')
    if np.any(null_msk):
        ax[0].scatter([0]*_n_null, _y_axis[null_msk], color=color, s=8)
        ax[0].vlines([0]*_n_null, ymin=ymin[null_msk], ymax=ymax[null_msk], color=color)
    ax[0].set_ylabel('bias')
    ax[0].get_xaxis().set_ticks([])
    ax[0].set_xlabel('Nulls')

    # x defined bias plot
    ax[1].axhline(0, linestyle='--', color='gray')
    ax[1].scatter(result.loc[~null_msk, 'x_mid'], _y_axis[~null_msk], color=color, s=8)
    ax[1].vlines(result.loc[~null_msk, 'x_mid'], ymin=ymin[~null_msk], ymax=ymax[~null_msk], color=color)
    ax[1].set_xlabel(x_col if x_col is not None else 'x')

    # categorical values
    if 'x_cat' in result.columns:
        ax[1].set_xticks(result.loc[~null_msk, 'x_mid'])
        ax[1].set_xticklabels(result.loc[~null_msk, 'x_cat'])

    plt.show()
    return ax


def analyze_feature(df: pd.DataFrame, x_col: str, p_col: str, y_col: str, w_col: Optional[str] = None,
                    segmentor: AbstractSegmentor = NtileSegmentor(),
                    ci_method: Optional[ConfidenceIntervalMethod] = BootstrapConfidenceIntervalMethod(),
                    plot: bool = True,
                    **plot_args) -> Tuple[pd.DataFrame, Optional[Axes]]:
    """
    Graphical test for the incremental predictive power of a feature over an existing probability model
    :param df: a pandas dataframe that contains columns that are (at least) the feature of interest,
        the prediction from a probabilistic model
        and the labels for analysis
    :param x_col: the name of the column with the feature (may contain Nans)
    :param p_col: the name of the column with the model predictions (no Nans allowed)
    :param y_col: the name of the column with the labels (no Nans allowed)
    :param w_col: an optional weight column, weights must be strictly positive
    :param segmentor: any AbstractSegmentor that will segment the feature, default is NtileSegmentor
    :param ci_method: any ConfidenceIntervalMethod, uses BootstrapConfidenceIntervalMethod by default
    :return: a dataframe with columns:
        i - denoting the segment's identifier, this would go from 0 to [num segments - 1], it will include a -1 for Nans
        x_min/x_max - the lowest/highest value of x in the segment
        bias_lower/bias_upper - according to the ci_method, these will be lower and upper confidence bounds of the bias (a)
        bias - the point estimate of the bias
        bis_std - only relevant for certain ci_methods
        n - the number of samples in the segment
        w - the sum of weights in the segments

    """
    x = df[x_col].values
    p = df[p_col].values
    y = df[y_col].values
    w = df[w_col] if w_col is not None else np.ones(df.shape[0])
    assert x.reshape(-1).size == x.shape[0]
    assert x.shape == y.shape
    assert p.shape == y.shape
    if w is None:
        w = np.ones(p.size)
    else:
        assert w.shape == y.shape
    assert ~np.isnan(p).any(), f"""no nans allowed in p"""
    assert ~np.isnan(w).any(), f"""no nans allowed in w"""
    assert ~np.isnan(y).any(), f"""no nans allowed in y"""

    xpyw = np.column_stack([x, p, y, w])
    s = segmentor.segment(xpyw, 0) # s has the the structure (x,p,y,w,segment)
    result = []

    for i in np.unique(s[:, -1]):
        s_i = s[s[:, -1] == i, :4] 
        r_i = {'i': i}
        lower, upper, avg, std, n = ci_method.ci(s_i[:, 1].astype(np.float64), 
                                                 s_i[:, 2].astype(np.float64), 
                                                 s_i[:, 3].astype(np.float64))
        if isinstance(segmentor, CategoricalSegmentor):
            r_i['x_cat'] = s_i[0, 0]
            r_i['x_mid'] = i
        else:
            r_i['x_min'] = s_i[:, 0].min()
            r_i['x_max'] = s_i[:, 0].max()
            r_i['x_mid'] = (r_i['x_min'] + r_i['x_max'])*0.5        

        r_i['bias_lower'] = lower
        r_i['bias_upper'] = upper
        r_i['bias'] = avg
        r_i['bias_std'] = std
        r_i['n'] = n
        r_i['w'] = s_i[:, 3].sum()
        r_i['p_q10'] = np.percentile(s_i[:, 1], 10)
        r_i['p_q25'] = np.percentile(s_i[:, 1], 25)
        r_i['p_q50'] = np.percentile(s_i[:, 1], 50)
        r_i['p_q75'] = np.percentile(s_i[:, 1], 75)
        r_i['p_q90'] = np.percentile(s_i[:, 1], 90)
           
        result.append(pd.Series(r_i))
    
    result = pd.DataFrame(result)
    if plot:
        ax = plot_result(result, **plot_args)

    return result, ax if plot else None


from typing import Union

from unittest import TestCase

from scipy.optimize import minimize
import numpy as np
from numpy.random import RandomState

from ifa.api import *
from ifa.math import sigmoid

def model_samples(n: int, beta: Union[int, np.ndarray], seed: Union[int, RandomState]=556342):
    """creates n random instances based on beta"""
    assert n > 0

    # resolve beta
    if isinstance(beta, int):
        beta = np.random.uniform(-1, 1, size=beta)
    elif isinstance(beta, np.ndarray):
        assert beta.size == beta.shape[0]
    else:
        raise TypeError()
    
    # resolve random state
    if isinstance(seed, int):
        rs = RandomState(seed)
    elif isinstance(seed, RandomState):
        rs = seed
    else:
        raise TypeError()
        
    p = beta.size

    x = rs.uniform(low=-1, high=1, size=(n, p))
    
    xb = x.dot(beta)
    p = sigmoid(xb)
    y = rs.binomial(1, p)
    
    return x, y, beta

def crossentropy_loss(x, X, y):
    mu = X.dot(x)
    y_hat = sigmoid(mu)
    return -(y*np.log(y_hat) + (1-y)*np.log(1-y_hat)).mean()

def fit_logistic_regression(X, y):
    opt = minimize(crossentropy_loss, np.zeros(X.shape[1]), (X, y))
    return opt['x']


class TestApi(TestCase):
   
    def test_example(self):
        k = 5
        rs = RandomState(333)
        beta = rs.uniform(-5,5,k)
        X, y, beta = model_samples(50000, beta) # missing 1 feature

        beta_hat = fit_logistic_regression(X, y)
        error = np.pow(beta - beta_hat,2).mean()
        print(f"beta error: {error}")

        X_partial = X[:,:(k-1)]
        beta_hat_partial = fit_logistic_regression(X_partial, y)
        p = sigmoid(X_partial.dot(beta_hat_partial))
        df = pd.DataFrame({'x':X[:,k-1], 'y': y, 'p': p})
        result, _ = analyze_feature(df, x_col='x', p_col='p', y_col='y', plot=False)
        result = result['bias']
        expected = np.array([ 1.114955,0.801960,0.627544,0.315205,0.136839,
                             -0.115860,-0.329878,-0.578883,-0.884712,-1.074554])
        np.testing.assert_allclose(result, expected, atol=1e-6, rtol=1e-6)

        


from typing import Union, Optional

import numpy as np
from scipy.optimize import minimize

def sigmoid(x: Union[np.ndarray, float]) -> Union[np.ndarray, float]:
    """the sigmoid function f(x)=\frac{1}{1+e^-x}
    
    Args:
        x - either a float or a numpy array of numeric values

    Returns:
        float or array, depending on the input
    """
    return 1/(1+np.exp(-x))


def clipped_probs(p: Union[np.ndarray, float], tau: float = 1e-6) -> Union[np.ndarray, float]:
    """
    Symmetrical clipping of probability vectors - 
    used to avoid numerical problems with logs of probablities close to 0,1
    
    Usage:
        >>> p = np.array([0, 1, 0.3])
        >>> clipped_probs(p, tau=0.1)
        returns: array([0.1, 0.9, 0.3])

    Args:
        p - vector with all values within [0,1] or a float within that range
        tau - the symmetric clipping value, must be a float in the range (0,0.5)

    Returns:
        either a vector or float depending on the input
    """
    assert 0 < tau < 0.5
    p = np.clip(p, a_min=tau, a_max=1-tau)
    return p


def logit(p: Union[np.ndarray, float], tau: float = 1e-6) -> Union[np.ndarray, float]:
    """
    The logit function

    Args:
        p - vector with all values within [0,1] or a float within that range
        tau - the symmetric clipping value, must be a float in the range (0,0.5)

    Returns:
        logit value as a float or vector, depending on the input

    """
    p_clipped = clipped_probs(p, tau)
    return np.log(p_clipped/(1-p_clipped))


def ce(y: np.ndarray, p: np.ndarray, w: np.ndarray, tau: float = 1e-6) -> float:
    """"(weighted) cross entropy
    
    Args:
        y - 1-d vector of {0,1} of size n, can also be within the range [0,1]
        p - 1-d vector of probabilities [0,1] of size n
        w - 1-d vector of weights, each w_i > 0 of size n
        tau - the symmetric clipping value, must be a float in the range (0,0.5)
    
    Returns:
        A float denoting the weighted cross entropy

    Raises:
        AssertionError:
          - if vectors are not 1-d
          - weights are not strictly positive
          - any p_i is not within [0,1]
          - any y_i is not within [0,1]    
    """
    assert y.shape[0] == y.reshape(-1).size, f'y needs to be 1 dimensional, got {y.shape=}'
    assert p.shape[0] == p.reshape(-1).size, f'p needs to be 1 dimensional, got {p.shape=}'
    assert w.shape[0] == w.reshape(-1).size, f'w needs to be 1 dimensional, got {w.shape=}'
    w_cond = w > 0
    assert w_cond.all(), f"""weights must be strictly positive, got values: {w[~w_cond]}"""
    p_cond = np.logical_and((0 <= p), (p <= 1))
    y_cond = np.logical_and((0 <= y), (y <= 1))
    assert p_cond.all(), f"""p must be within (0,1), got values: {p[~p_cond]}"""
    assert y_cond.all(), f"""y must be within [0,1], got values: {y[~y_cond]}"""

    p = clipped_probs(p, tau)
    loss = y*np.log(p) + (1-y)*np.log(1-p)
    loss *= w
    return -loss.sum() / w.sum()


def objective(a: float, y: np.ndarray, p: np.ndarray, 
              w: Optional[np.ndarray] = None, tau: float = 1e-6) -> float:
    """IFA's classification objective
     This is parameterized CE with 1 global additive scalar on the logit scale
     
     See documentation for the mathematical details

     Args:
        a - float, the global additive scalar to the CE loss
        y - 1-d vector of {0,1} of size n, can also be within the range [0,1]
        p - 1-d vector of probabilities [0,1] of size n
        w - 1-d vector of weights, each w_i > 0 of size n
        tau - the symmetric clipping value, must be a float in the range (0,0.5)

    Returns:
        a float denoting the CE adjusted by a

    Raises:
        see the documentation for the ifa.math.ce method
     
     """
    if w is None:
        w = np.ones(y.size)
    p_a = sigmoid(logit(p) + a)
    return ce(y, p_a, w, tau)


def minimize_ce_with_1param(p: np.ndarray, y: np.ndarray, w: Optional[np.ndarray] = None, 
                            verbose: bool = False) -> float:
    """given a structure (p,y,w) finds the constant a that will minimize CE
    for a given observation i, the unweighted CE is defined as:
        y_i * log(\sigma(\sigma^{-1}(p_i)+a)) + (1-y_i) * log(1-\sigma(\sigma^{-1}(p_i)+a)))

    Args:
        y - 1-d vector of {0,1} of size n, can also be within the range [0,1]
        p - 1-d vector of probabilities [0,1] of size n
        w - 1-d vector of weights, each w_i > 0 of size n
        verbose - bool, will print the scipy.optimize.OptimizeResult object returned from calling scipy.optimize.mininize
    
    Returns:
        a float denoting the arg that will minimize CE as an added constant in logit scale
    """
    assert p.shape == y.shape

    opt = minimize(fun=objective, x0=np.array([0.0]), args=(y, p, w))
    if verbose:
        print(opt)
    return opt['x'][0]


from unittest import TestCase


from ifa.segmentor import *

class TestSegmentor(TestCase):

    def test_segmentor(self):
        rs = np.random.RandomState(42)
        x = rs.uniform(0, 1, 100).reshape(-1,1)
        segmentor = NtileSegmentor(5)
        self.assertTrue(segmentor.state is None)
        s = segmentor.segment(x, 0)
        self.assertEqual(s.shape, (x.size, 2))
        self.assertTrue(set(np.unique(s[:, 1]).tolist()) == {1., 2., 3., 4., 5.})
        self.assertTrue(np.abs(segmentor.state - np.array([0.00552212, 0.15601382, 0.32867095, 0.57373215, 0.77282238, 0.98688694])).mean() < 1e-8)

        segmentor = NtileSegmentor(10)
        self.assertTrue(segmentor.state is None)
        s = segmentor.segment(x, 0)
        self.assertEqual(s.shape, (x.size, 2))
        self.assertTrue(set(np.unique(s[:, 1]).tolist()) == {1., 2., 3., 4., 5., 6., 7., 8., 9., 10.})
        self.assertTrue(np.abs(segmentor.state - 
                               np.array([0.00552212, 0.07450004, 0.15601382, 0.26757832, 0.32867095, 0.46414245,
                                         0.57373215, 0.69102032, 0.77282238, 0.8879742,  0.98688694])).mean() < 1e-8)

        x[:10] = np.nan
        segmentor = NtileSegmentor(3)
        self.assertTrue(segmentor.state is None)
        s = segmentor.segment(x, 0)
        self.assertEqual(s.shape, (x.size, 2))
        self.assertEqual(len(np.unique(s[:, 1])), 4)
        self.assertTrue(np.all(np.isnan(s[s[:,1] == -1][:,0])))
        self.assertTrue(np.abs(segmentor.state - np.array([0.00552212, 0.29183948, 0.61566797, 0.98688694])).mean() < 
        1e-8)

        x = np.array(['a', 'b', 'c', 'c', np.nan, None]).reshape(-1,1)
        segmentor = CategoricalSegmentor()
        self.assertTrue(segmentor.state is None)
        s = segmentor.segment(x, 0)
        self.assertTrue(s.shape == (6,2))
        self.assertListEqual(s[:,0].tolist(), x.reshape(-1).tolist())
        self.assertEqual(np.unique(s[:,1]).size, 4)
        self.assertDictEqual({'a': 2, 'b': 3, 'c': 1}, segmentor.state)

        # segmentor with top_k = 2
        segmentor = CategoricalSegmentor(2)
        self.assertTrue(segmentor.state is None)
        s = segmentor.segment(x, 0)
        self.assertTrue(s.shape == (6,2))
        self.assertEqual(np.unique(s[:,1]).size, 4)
        self.assertTrue('_other_' in s[:,0].tolist())
        self.assertDictEqual({'a': 2, 'c': 1}, segmentor.state)

        # segmentor with top_k = 1
        segmentor = CategoricalSegmentor(1)
        s = segmentor.segment(x, 0)
        self.assertTrue(s.shape == (6,2))
        self.assertEqual(np.unique(s[:,1]).size, 3)
        self.assertTrue('_other_' in s[:,0].tolist())

        # segmentor with top_k = 1, and other name
        segmentor = CategoricalSegmentor(1, 'xxx')
        s = segmentor.segment(x, 0)
        self.assertTrue(s.shape == (6,2))
        self.assertEqual(np.unique(s[:,1]).size, 3)
        self.assertTrue('xxx' in s[:,0].tolist())


from abc import ABC, abstractmethod
from typing import Optional
from collections import Counter

import numpy as np

class AbstractSegmentor(ABC):
    """
    IFA segments the data into groups in order to assess the potential improvement in CE in that area
    The AbstractSegmentor is the abstract segmentor class, any concrete class needs to 
    implement the _fit and _segment_implementation methods.
    """

    def __init__(self):
        self.state = None
    
    def segment(self, v: np.ndarray, _i: int) -> np.ndarray:
        """
        Args:
            v - is a k dimensional array.
            _i - is the index that we want to segment by.
        Returns:
            a k+1 dimensional array with:
                - original v in the first k columns
                - a mapping of x to segments in the k+1 column
                Values of np.nans will be mapped into -1"""
        assert 0 <= _i <= v.shape[1] - 1
        n, k = v.shape
        mask = ~np.isnan(v[:,_i])
        xnn = v[mask, :]
        if self.state is None:
            self._fit(xnn, _i)
        
        no_nan_result = self._segment_implementation(xnn, _i)
        assert no_nan_result.shape[1] == k + 1, f'num of cols in no_nan_result={no_nan_result.shape[1]}, ' \
                                                f'expecting {k+1}'
        assert no_nan_result.shape[0] == mask.sum()
        nan_result = v[~mask, :]
        nan_result = np.column_stack([nan_result, -np.ones(n-mask.sum())])
        result = np.row_stack([no_nan_result, nan_result])

        return result

    @abstractmethod
    def _fit(self, x: np.ndarray, i: int):
        pass

    @abstractmethod
    def _segment_implementation(self, x: np.ndarray, i: int) -> np.ndarray:
        pass


class CategoricalSegmentor(AbstractSegmentor):

    def __init__(self, top_k: Optional[int]=None, other_category_name: str='_other_'):
        super().__init__()
        if top_k is not None:
            assert top_k > 0, f'`top_k` must be greater than 0 or None, got {top_k=}'
        self.top_k = top_k
        self.other_category_name = other_category_name

    @staticmethod
    def is_nan(v):
        if v is None:
            return True
        elif isinstance(v, (str,)):
            return v == 'nan'
        elif isinstance(v, (float, int)):
            return np.isnan(v)
        else:
            raise ValueError(f'unrecognized type: {v=}, {type(v)=}')

    def segment(self, v: np.ndarray, _i: int) -> np.ndarray:
        nan_mask = np.array([self.is_nan(v_i) for v_i in v[:, _i].tolist()])
        v_nan = v[nan_mask,:]
        n_nan = v_nan.shape[0]
        v_no_nan = v[~nan_mask,:]
        if self.state is None:
            self._fit(v_no_nan, _i)
        
        def _class(x_i):
            if x_i in self.state:
                return self.state[x_i]
            elif x_i is None:
                return -1
            else:
                return 0 # other
            
        s = v_no_nan[:, _i].tolist()
        ss = np.array(list(map(_class, s)))
        ret = np.column_stack([v_no_nan, ss])
        ret_nan = np.column_stack([v_nan, [-1]*n_nan])
        result = np.row_stack([ret, ret_nan])
        result[result[:,-1] == 0, 0] = self.other_category_name
        return result

    def _segment_implementation(self, x: np.ndarray, i: int) -> np.ndarray:
        pass

    def _fit(self, x: np.ndarray, i: int):
        c = Counter(x[:, i])
        c = sorted(c.items(), key=lambda t: -t[1])
        if self.top_k is not None:    
            c = c[:self.top_k]
            c = list(map(lambda t: t[0], c))
        else:
            c = [k for k,v in c]

        self.state = {k: j+1 for j, k in enumerate(c)}
        

class NtileSegmentor(AbstractSegmentor):

    def __init__(self, tiles: int = 10):
        super().__init__()
        assert tiles > 0
        self.tiles = tiles

    def _fit(self, array: np.ndarray, i: int):
        percentiles = np.linspace(0, 100, self.tiles + 1)
        self.state = np.percentile(array[:, i], percentiles)

    def _segment_implementation(self, array: np.ndarray, i: int):
        tile_values = self.state
        mapped_array = np.zeros((array.shape[0], array.shape[1]+1))
        mapped_array[:, :array.shape[1]] = array

        for j in range(1, self.tiles + 1):
            indices = np.logical_and(array[:, i] >= tile_values[j - 1], array[:, i] <= tile_values[j])
            mapped_array[indices, array.shape[1]] = j

        return mapped_array


from abc import ABC, abstractmethod
from typing import Optional, Tuple, Callable, Union

import scipy
import numpy as np
from numpy.random import RandomState

from ifa.math import objective, minimize_ce_with_1param

class ConfidenceIntervalMethod(ABC):

    def __init__(self, alpha: float = 0.05):
        assert 0 < alpha < 1
        self.alpha = alpha

    def ci(self, p: np.ndarray, y: np.ndarray, w: Optional[np.ndarray] = None) -> \
            Tuple[float, float, float, float, int]:
        """returns a tuple (lower, upper, avg, stddev, n)"""
        assert p.reshape(-1).size == p.shape[0]
        assert p.shape == y.shape
        if w is None:
            w = np.ones(p.size)
        else:
            assert w.shape == y.shape
        assert ~np.isnan(p).any(), f"""no nans allowed in p"""
        assert ~np.isnan(w).any(), f"""no nans allowed in w"""
        assert ~np.isnan(y).any(), f"""no nans allowed in y"""

        return self._ci_implementation(p, y, w)

    @abstractmethod
    def _ci_implementation(self, p: np.ndarray, y: np.ndarray, w: Optional[np.ndarray] = None) -> \
            Tuple[float, float, float, float, int]:
        pass


class PointEstimateMethod(ConfidenceIntervalMethod):

    def _ci_implementation(self, p: np.ndarray, y: np.ndarray, w: Optional[np.ndarray] = None) -> \
            Tuple[float, float, float, float, int]:
        n = p.size
        a = minimize_ce_with_1param(p, y, w)
        return a, a, a, np.nan, n


class LikelihoodRatioConfidenceIntervalMethod(ConfidenceIntervalMethod):

    def __init__(self, alpha: float = 0.05,
                 tol: float = 1e-10,
                 max_iter: Optional[int] = None,
                 verbose: bool = False):
        super().__init__(alpha)
        assert 0 < tol < 0.1
        self.tol = tol
        self.max_iter = max_iter
        self.verbose = verbose

    @staticmethod
    def log_likelihood_ratio(a_opt: float, p: np.ndarray, y: np.ndarray, w: Optional[np.ndarray] = None):
        def f(a: float):
            ratio = objective(a, y, p, w) / objective(a_opt, y, p, w)
            llr = -2*np.log(ratio)
            return -llr if a >= a_opt else llr
        return f

    @staticmethod
    def bisection(target: float, f: Callable[[float], float],
                  x_lo: float, x_hi: float, tol: float = 1e-10,
                  max_iter: Optional[int] = 100,
                  verbose: bool = False):
        i = 0
        while abs(x_lo - x_hi) > tol and (True if max_iter is None else i <= max_iter):
            if verbose:
                print(f'bisection {i=}, {x_lo=}, {x_hi=}, diff={abs(x_lo - x_hi)}')
            ad = abs(x_lo - x_hi)
            f_lo = f(x_lo)
            f_hi = f(x_hi)
            if target <= f_lo:
                x_hi = x_lo
                x_lo = x_lo - ad
            elif f_lo < target < f_hi:
                x_lo = (x_lo + x_hi)*0.5
            else:
                x_lo = x_hi
                x_hi = x_lo + ad
            i += 1
        return (x_lo + x_hi)*0.5

    def _ci_implementation(self, p: np.ndarray, y: np.ndarray, w: Optional[np.ndarray] = None) -> \
            Tuple[float, float, float, float, int]:
        a_opt = minimize_ce_with_1param(p, y, w)
        chi2 = scipy.stats.chi2(1)
        chi2_q = chi2.ppf(1-self.alpha)
        llr = self.log_likelihood_ratio(a_opt, p, y, w)

        upper = self.bisection(chi2_q, llr, a_opt, a_opt+4, self.tol, max_iter=self.max_iter, verbose=self.verbose)
        lower = self.bisection(-chi2_q, llr, a_opt, a_opt+4, self.tol, max_iter=self.max_iter, verbose=self.verbose)

        return lower, upper, a_opt, np.nan, p.size


class BootstrapConfidenceIntervalMethod(ConfidenceIntervalMethod):

    def __init__(self, alpha: float = 0.05, n_samples: int = 10, seed: Union[int, RandomState] = 4242):
        super().__init__(alpha)
        assert isinstance(n_samples, int)
        assert n_samples > 0
        self.n_samples = n_samples
        self.rs = self._resolve_random_state(seed)

    @staticmethod
    def _resolve_random_state(seed: int = Union[int, RandomState]) -> RandomState:
        if isinstance(seed, int):
            return RandomState(seed)
        elif isinstance(seed, RandomState):
            return seed
        else:
            raise TypeError(f'seed needs to be either [int, RandomState], got {str(type(seed))}')

    def bootstrap_indices(self, n: int):
        return self.rs.choice(range(n), size=(n, self.n_samples), replace=True)

    def bootstrap_results(self, p: np.ndarray, y: np.ndarray, w: np.ndarray):
        pyw = np.column_stack([p, y, w])
        n = pyw.shape[0]
        results = []

        indices = self.bootstrap_indices(n)

        for i in range(self.n_samples):
            sample_i = pyw[indices[:, i], :]
            _p, _y, _w = sample_i[:, 0], sample_i[:, 1], sample_i[:, 2]
            a = minimize_ce_with_1param(_p, _y, _w)
            results.append(a)

        return np.array([results])

    def _ci_implementation(self, p: np.ndarray, y: np.ndarray, w: Optional[np.ndarray] = None) -> \
            Tuple[float, float, float, float, int]:

        results = self.bootstrap_results(p, y, w)
        lower = np.percentile(results, 100*self.alpha/2)
        upper = np.percentile(results, 100*(1-self.alpha/2))
        avg = results.mean()
        std = results.std()
        return lower, upper, avg, std, p.size


from setuptools import setup, find_packages

setup(
    name="ak_ifa",  
    version="0.1.0",
    author="Aviad Klein",
    author_email="aviad.klein@gmail.com",
    description="Incremental Feature Analysis Tool",
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url="https://github.com/AviadKlein/ifa",  
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.11',
)


from unittest import TestCase

from ifa.math import *
import numpy as np

class TestMath(TestCase):

    def assert_arrays_allclose(self,a,b,tol=1e-20):
        np.testing.assert_allclose(a,b,tol,tol)

    def test_math(self):
        self.assertEqual(sigmoid(0), 0.5)
        self.assertTrue(sigmoid(-1) < sigmoid(0))
        self.assertTrue(sigmoid(-100) < sigmoid(-1))
        self.assertTrue(sigmoid(-100) < 1e-10)

        self.assertEqual(clipped_probs(0.99, tau=0.1), 0.90)
        self.assertEqual(clipped_probs(0.01, tau=0.1), 0.1)

        self.assertEqual(logit(0.5), 0)
        self.assertTrue(logit(0.7) > logit(0.6) > logit(0.5))
        self.assertEqual(logit(0.8), logit(0.8, tau=0.2))

        with self.assertRaises(AssertionError) as cm:
            y = np.random.uniform(0,1,size=(10, 2))
            ce(y, y, y)
        self.assertTrue('y needs to be 1 dimensional, got y.shape=(10, 2)' in str(cm.exception))


    def test_sigmoid(self):
        result = sigmoid(0)
        self.assertEqual(result, 0.5)

        x = np.array([-100, -1, 1, 100])
        result = sigmoid(x)
        expected = [0, 1/(1+np.exp(1)), 1/(1+np.exp(-1)), 1]
        self.assert_arrays_allclose(result, expected, 1e-20)

    def test_clipped_probs(self):
        p = np.array([0, 1, 0.3])
        result = clipped_probs(p, tau=0.1)
        expected = np.array([0.1, 0.9, 0.3])
        self.assert_arrays_allclose(result, expected, 1e-20)

    def test_logit(self):
        p = np.array([0.2, 0.5, 0.6])
        result = logit(p)
        expected = np.array([-1.38629436, 0., 0.40546511])
        self.assert_arrays_allclose(result, expected, 1e-8)

    def test_ce_unweighted_single(self):
        y = np.array([0.2])
        p = np.array([0.2])
        w = np.ones(1)
        result = ce(y,p,w)
        expected = -(0.2*np.log(0.2) + 0.8*np.log(0.8))
        self.assert_arrays_allclose(result, expected)

    def test_ce_weighted_single(self):
        y = np.array([0.2])
        p = np.array([0.2])
        w = np.ones(1)*5
        result = ce(y,p,w)
        expected = -(0.2*np.log(0.2) + 0.8*np.log(0.8))
        self.assert_arrays_allclose(result, expected)

    def test_ce_unweighted_multiple(self):
        y = np.array([0.2, 0.8])
        ps = [np.array([0.2, _]) for _ in [0.7, 0.8, 0.9]]
        result = [ce(y, ps_i, np.ones(2)) for ps_i in ps]
        self.assertEqual(result[1], min(result))
    
    def test_objective_a_is_zero(self):
        y = np.array([0, 1, 1])
        p = np.array([0.1, 0.3, 0.3])
        result = objective(0, y, p, np.ones(3))
        expected = ce(y, p, np.ones(3))
        self.assertAlmostEqual(result, expected, 6)

    def test_objective(self):
        y = np.array([0, 1, 1])
        p = np.array([0.1, 0.3, 0.3])
        ce_result = ce(y, p, np.ones(3))
        self.assertTrue(objective(-0.1, y, p, np.ones(3)) > ce_result)
        self.assertTrue(objective(0.1, y, p, np.ones(3)) < ce_result)

    def test_minimizer(self):
        y = np.array([0, 1, 1])
        p = np.array([0.1, 0.3, 0.3])
        result = minimize_ce_with_1param(p, y, np.ones(3))
        orig = ce(p, y, np.ones(3))
        minima = objective(result, y, p, np.ones(3))
        self.assertTrue(minima < orig)



        


